@@ -1,5 +1,7 @@
 #!/usr/bin/env python
 
+__version__ = '0.6.5'
+
 import base64
 import datetime
 import hashlib
@@ -65,8 +67,8 @@ else:
         return highlight(data, lexer, formatter)
 
 try:
-    from peewee import __version__
-    peewee_version = tuple([int(p) for p in __version__.split('.')])
+    from peewee import __version__ as _pw_version
+    peewee_version = tuple([int(p) for p in _pw_version.split('.')])
 except ImportError:
     raise RuntimeError('Unable to import peewee module. Install by running '
                        'pip install peewee')
@@ -74,7 +76,7 @@ else:
     if peewee_version < (3, 0, 0):
         raise RuntimeError('Peewee >= 3.0.0 is required. Found version %s. '
                            'Please update by running pip install --update '
-                           'peewee' % __version__)
+                           'peewee' % _pw_version)
 
 from peewee import *
 from peewee import IndexMetadata
@@ -85,8 +87,9 @@ from playhouse.migrate import migrate
 
 CUR_DIR = os.path.realpath(os.path.dirname(__file__))
 DEBUG = False
-MAX_RESULT_SIZE = 1000
 ROWS_PER_PAGE = 50
+QUERY_ROWS_PER_PAGE = 1000
+TRUNCATE_VALUES = True
 SECRET_KEY = 'sqlite-database-browser-0.1.0'
 
 app = Flask(
@@ -274,6 +277,7 @@ def _query_view(template, table=None):
     data = []
     data_description = error = row_count = sql = None
     ordering = None
+    pk_index = None
 
     sql = qsql = request.values.get('sql') or ''
 
@@ -302,33 +306,80 @@ def _query_view(template, table=None):
         default_sql = ''
         model_class = dataset._base_model
 
+    page = page_next = page_prev = page_start = page_end = total_pages = 1
+    total = -1  # Number of rows in query result.
     if qsql:
         if export_format:
             query = model_class.raw(qsql).dicts()
             return export(query, export_format, table)
 
+        try:
+            total, = dataset.query('SELECT COUNT(*) FROM (%s) as _' %
+                                   qsql.rstrip('; ')).fetchone()
+        except Exception as exc:
+            total = -1
+
+        # Apply pagination.
+        rpp = app.config['QUERY_ROWS_PER_PAGE']
+        page = request.values.get('page')
+        if page and page.isdigit():
+            page = max(int(page), 1)
+        else:
+            page = 1
+        offset = (page - 1) * rpp
+        page_prev, page_next = max(page - 1, 1), page + 1
+
+        # Figure out highest page.
+        if total > 0:
+            total_pages = max(1, int(math.ceil(total / float(rpp))))
+            page = max(min(page, total_pages), 1)
+            page_next = min(page + 1, total_pages)
+            page_start = offset + 1
+            page_end = min(total, page_start + rpp - 1)
+
+        if page > 1:
+            qsql = ('SELECT * FROM (%s) AS _ LIMIT %d OFFSET %d' %
+                    (qsql.rstrip(' ;'), rpp, offset))
+
         try:
             cursor = dataset.query(qsql)
         except Exception as exc:
             error = str(exc)
             app.logger.exception('Error in user-submitted query.')
         else:
-            data = cursor.fetchall()[:app.config['MAX_RESULT_SIZE']]
+            data = cursor.fetchmany(rpp)
             data_description = cursor.description
             row_count = cursor.rowcount
 
+    if data_description is not None and table:
+        col_names = [r[0] for r in data_description]
+        pk = model_class._meta.primary_key
+        if not isinstance(pk, CompositeKey) and pk is not False:
+            pk = pk.column_name
+            if pk in col_names:
+                pk_index = col_names.index(pk)
+
     return render_template(
         template,
         data=data,
         data_description=data_description,
         default_sql=default_sql,
         error=error,
+        is_view=any(v.name == table for v in dataset.get_all_views()),
         ordering=ordering,
+        page=page,
+        page_end=page_end,
+        page_next=page_next,
+        page_prev=page_prev,
+        page_start=page_start,
+        pk_index=pk_index,
         query_images=get_query_images(),
         row_count=row_count,
         sql=sql,
         table=table,
-        table_sql=dataset.get_table_sql(table))
+        table_sql=dataset.get_table_sql(table),
+        total=total,
+        total_pages=total_pages)
 
 @app.route('/query/', methods=['GET', 'POST'])
 def generic_query():
@@ -348,8 +399,7 @@ def table_create():
     if not table:
         flash('Table name is required.', 'danger')
         dest = request.form.get('redirect') or url_for('index')
-        if not dest.startswith('/'):
-            dest = '/' + dest
+        dest = '/' + dest.lstrip('/')  # idiot vulnerability "researchers".
         return redirect(dest)
 
     try:
@@ -404,6 +454,7 @@ def add_column(table):
     name = request_data.get('name', '')
 
     if request.method == 'POST':
+        name = re.sub(r'[^\w]+', '_', name.strip())
         if name and col_type in column_mapping:
             try:
                 migrate(
@@ -471,6 +522,7 @@ def rename_column(table):
     column_names = [column.name for column in columns]
 
     if request.method == 'POST':
+        rename_to = re.sub(r'[^\w]+', '_', rename_to.strip())
         if (rename in column_names) and (rename_to not in column_names):
             try:
                 migrate(migrator.rename_column(table, rename, rename_to))
@@ -599,8 +651,7 @@ def table_content(table):
     rows_per_page = app.config['ROWS_PER_PAGE']
     total_pages = max(1, int(math.ceil(total_rows / float(rows_per_page))))
     # Restrict bounds.
-    page_number = min(page_number, total_pages)
-    page_number = max(page_number, 1)
+    page_number = max(min(page_number, total_pages), 1)
 
     previous_page = page_number - 1 if page_number > 1 else None
     next_page = page_number + 1 if page_number < total_pages else None
@@ -642,8 +693,11 @@ def minimal_validate_field(field, value):
         return 'NULL', 'Column does not allow NULL values.'
     if value is None:
         return None, None
-    if isinstance(field, IntegerField) and not value.isdigit():
-        return value, 'Value is not a number.'
+    if isinstance(field, IntegerField):
+        try:
+            _ = int(value)
+        except Exception:
+            return value, 'Value is not a number.'
     elif isinstance(field, FloatField):
         try:
             _ = float(value)
@@ -672,27 +726,29 @@ def table_insert(table):
     model = dataset[table].model_class
 
     columns = []
-    col_dict = {}
+    fields = []
+    defaults = {}
     row = {}
     for column in dataset.get_columns(table):
         field = model._meta.columns[column.name]
         if isinstance(field, AutoField):
             continue
+        if column.default:
+            defaults[column.name] = column.default
         columns.append(column)
-        col_dict[column.name] = column
-        row[column.name] = ''
+        fields.append(field)
+        row[field.name] = ''
 
     edited = set()
     errors = {}
     if request.method == 'POST':
         insert = {}
         for key, value in request.form.items():
-            if key not in col_dict: continue
-            column = col_dict[key]
-            edited.add(column.name)
-            row[column.name] = value
+            if key not in model._meta.fields: continue
+            field = model._meta.fields[key]
+            edited.add(field.name)
+            row[field.name] = value
 
-            field = model._meta.columns[column.name]
             value, err = minimal_validate_field(field, value)
             if err:
                 errors[key] = err
@@ -718,11 +774,14 @@ def table_insert(table):
         else:
             flash('No data was specified to be inserted.', 'warning')
     else:
-        edited = set(col_dict)  # Make all fields editable on load.
+        edited = set(model._meta.sorted_field_names) - set(defaults)  # Make all fields editable on load.
+
+    columns_fields = zip(columns, fields)
 
     return render_template(
         'table_insert.html',
-        columns=columns,
+        columns_fields=columns_fields,
+        defaults=defaults,
         edited=edited,
         errors=errors,
         model=model,
@@ -763,30 +822,27 @@ def table_update(table, pk):
         return redirect(url_for('table_content', table=table))
 
     columns = dataset.get_columns(table)
-    col_dict = {}
+    fields = model._meta.sorted_fields
     row = {}
-    for column in columns:
-        value = getattr(obj, column.name)
+    for field in fields:
+        value = getattr(obj, field.name)
         if value is None:
-            row[column.name] = None
-        elif column.data_type.lower() == 'blob':
-            row[column.name] = base64.b64encode(value).decode('utf8')
+            row[field.name] = None
+        elif isinstance(field, BlobField):
+            row[field.name] = base64.b64encode(value).decode('utf8')
         else:
-            row[column.name] = value
-
-        col_dict[column.name] = column
+            row[field.name] = value
 
     edited = set()
     errors = {}
     if request.method == 'POST':
         update = {}
         for key, value in request.form.items():
-            if key not in col_dict: continue
-            column = col_dict[key]
-            edited.add(column.name)
-            row[column.name] = value
+            if key not in model._meta.fields: continue
+            field = model._meta.fields[key]
+            edited.add(field.name)
+            row[field.name] = value
 
-            field = model._meta.columns[column.name]
             value, err = minimal_validate_field(field, value)
             if err:
                 errors[key] = err
@@ -809,11 +865,14 @@ def table_update(table, pk):
         else:
             flash('No data was specified to be updated.', 'warning')
 
+    columns_fields = zip(columns, fields)
+
     return render_template(
         'table_update.html',
-        columns=columns,
+        columns_fields=columns_fields,
         edited=edited,
         errors=errors,
+        fields=fields,
         model=model,
         pk=pk,
         row=row,
@@ -854,7 +913,6 @@ def table_delete(table, pk):
 
     return render_template(
         'table_delete.html',
-        column_names=[c.name for c in dataset.get_columns(table)],
         model=model,
         pk=pk,
         row=row,
@@ -871,7 +929,7 @@ def export(query, export_format, table=None):
     if export_format == 'json':
         kwargs = {'indent': 2}
         filename = 'export.json'
-        mimetype = 'text/javascript'
+        mimetype = 'application/json'
     else:
         kwargs = {}
         filename = 'export.csv'
@@ -1046,10 +1104,13 @@ def value_filter(value, max_length=50):
     if isinstance(value, binary_types):
         if not isinstance(value, (bytes, bytearray)):
             value = bytes(value)  # Handle `buffer` type.
-        value = base64.b64encode(value)[:1024].decode('utf8')
+        try:
+            value = value.decode('utf8')
+        except UnicodeDecodeError:
+            value = base64.b64encode(value)[:1024].decode('utf8')
     if isinstance(value, unicode_type):
         value = escape(value)
-        if len(value) > max_length:
+        if len(value) > max_length and app.config['TRUNCATE_VALUES']:
             return ('<span class="truncated">%s</span> '
                     '<span class="full" style="display:none;">%s</span>'
                     '<a class="toggle-value" href="#">...</a>') % (
@@ -1100,6 +1161,7 @@ def _general():
     return {
         'dataset': dataset,
         'login_required': bool(app.config.get('PASSWORD')),
+        'version': __version__,
     }
 
 @app.context_processor
@@ -1182,13 +1244,35 @@ def get_option_parser():
         '--rows-per-page',
         default=50,
         dest='rows_per_page',
-        help='Number of rows to display per page (default=50)',
+        help='Number of rows to display per page in content tab (default=50)',
         type='int')
+    parser.add_option(
+        '-Q',
+        '--query-rows-per-page',
+        default=1000,
+        dest='query_rows_per_page',
+        help='Number of rows to display per page in query tab (default=1000)',
+        type='int')
+    parser.add_option(
+        '-T',
+        '--no-truncate',
+        action='store_false',
+        default=True,
+        dest='truncate_values',
+        help=('Disable truncating long text values. By default text values '
+              'are ellipsized after 50 characters and the full text is shown '
+              'on click.'))
     parser.add_option(
         '-u',
         '--url-prefix',
         dest='url_prefix',
         help='URL prefix for application.')
+    parser.add_option(
+        '-f',
+        '--foreign-keys',
+        action='store_true',
+        dest='foreign_keys',
+        help='Enable the foreign_keys pragma.')
     parser.add_option(
         '-e',
         '--extension',
@@ -1243,7 +1327,7 @@ def install_auth_handler(password):
             return redirect(url_for('login'))
 
 def initialize_app(filename, read_only=False, password=None, url_prefix=None,
-                   extensions=None):
+                   extensions=None, foreign_keys=None):
     global dataset
     global migrator
 
@@ -1269,6 +1353,9 @@ def initialize_app(filename, read_only=False, password=None, url_prefix=None,
     else:
         db = SqliteDatabase(filename)
 
+    if foreign_keys:
+        db.pragma('foreign_keys', True, permanent=True)
+
     if extensions:
         # Load extensions before performing introspection.
         for ext in extensions:
@@ -1296,6 +1383,10 @@ def main():
         handler.setFormatter(fmt)
         app.logger.addHandler(handler)
 
+        # Remove default handler.
+        from flask.logging import default_handler
+        app.logger.removeHandler(default_handler)
+
     password = None
     if options.prompt_password:
         if os.environ.get('SQLITE_WEB_PASSWORD'):
@@ -1311,10 +1402,14 @@ def main():
 
     if options.rows_per_page:
         app.config['ROWS_PER_PAGE'] = options.rows_per_page
+    if options.query_rows_per_page:
+        app.config['QUERY_ROWS_PER_PAGE'] = options.query_rows_per_page
+
+    app.config['TRUNCATE_VALUES'] = options.truncate_values
 
     # Initialize the dataset instance and (optionally) authentication handler.
     initialize_app(args[0], options.read_only, password, options.url_prefix,
-                   options.extensions)
+                   options.extensions, options.foreign_keys)
 
     if options.browser:
         open_browser_tab(options.host, options.port)
